"""
TABI - UCSC iGEM 2023

StealthGenome: Prep Genome for Stealth and Codon Frequency analysis

PlasmidParser: Identifies Stealth-mutable regions of an annotated plasmid

Author: Tyler Gaw
"""

from ._SeqParser import *


"""
TABI - UCSC iGEM 2023

FastA Reader class to read FastA files

Modified from FastAreader written by David L. Bernick

Author: David L. Bernick
"""

# Args:
#     fname (str): file name (optional), default is None (STDIN)

#     Usage:
#     thisReader = FastAreader ('testTiny.fa')
#     for head, seq in thisReader.readFasta():
#         print (head,seq)

from ._FastAreader import *


import chameleontools.FastAreader as FastAreader
import chameleontools.SeqParser as SeqParser
import Bio.SeqRecord
import random

class InvalidAA(KeyError):
    def __init__(self, message) -> None:
        super().__init__(message)

class CDSanalyzer:
    """
    Analyzes CDS regions 
    """
    
    dnaCodonTable = {
        # RNA codon table
        # T
        'TTT': 'F', 'TCT': 'S', 'TAT': 'Y', 'TGT': 'C',  # TxT
        'TTC': 'F', 'TCC': 'S', 'TAC': 'Y', 'TGC': 'C',  # TxC
        'TTA': 'L', 'TCA': 'S', 'TAA': '*', 'TGA': '*',  # TxA
        'TTG': 'L', 'TCG': 'S', 'TAG': '*', 'TGG': 'W',  # TxG
        # C
        'CTT': 'L', 'CCT': 'P', 'CAT': 'H', 'CGT': 'R',  # CxT
        'CTC': 'L', 'CCC': 'P', 'CAC': 'H', 'CGC': 'R',  # CxC
        'CTA': 'L', 'CCA': 'P', 'CAA': 'Q', 'CGA': 'R',  # CxA
        'CTG': 'L', 'CCG': 'P', 'CAG': 'Q', 'CGG': 'R',  # CxG
        # A
        'ATT': 'I', 'ACT': 'T', 'AAT': 'N', 'AGT': 'S',  # AxT
        'ATC': 'I', 'ACC': 'T', 'AAC': 'N', 'AGC': 'S',  # AxC
        'ATA': 'I', 'ACA': 'T', 'AAA': 'K', 'AGA': 'R',  # AxA
        'ATG': 'M', 'ACG': 'T', 'AAG': 'K', 'AGG': 'R',  # AxG
        # G
        'GTT': 'V', 'GCT': 'A', 'GAT': 'D', 'GGT': 'G',  # GxT
        'GTC': 'V', 'GCC': 'A', 'GAC': 'D', 'GGC': 'G',  # GxC
        'GTA': 'V', 'GCA': 'A', 'GAA': 'E', 'GGA': 'G',  # GxA
        'GTG': 'V', 'GCG': 'A', 'GAG': 'E', 'GGG': 'G'  # GxG
    }
    
    validAA = set(dnaCodonTable.values())
    
    def __init__(self, input: str | SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord]) -> None:
        """
        self.addCDS() -> adds a CDS for analysis
        self.getUsage() -> dictionary of codon usage grouped by AA translation | (optional) gathers codon usage by AA
        self.getFrequency() -> dictionary of codon frequency grouped by AA translation | (optional) gathers codon frequency by AA
        self.len() -> total length of analyzed CDS regions
        """
        self.codonUsage = {
            "F": {"TTT":0, "TTC":0},
            "L": {"TTA":0, "TTG":0, "CTT":0, "CTC":0, "CTA":0, "CTG":0},
            "S": {"TCT":0, "TCC":0, "TCA":0, "TCG":0, "AGT":0, "AGC":0},
            "Y": {"TAT":0, "TAC":0},
            "C": {"TGT":0, "TGC":0},
            "W": {"TGG":0},
            "P": {"CCT":0, "CCC":0, "CCA":0, "CCG":0},
            "H": {"CAT":0, "CAC":0},
            "Q": {"CAA":0, "CAG":0},
            "R": {"CGT":0, "CGC":0, "CGA":0, "CGG":0, "AGA":0, "AGG":0},
            "I": {"ATT":0, "ATC":0, "ATA":0},
            "M": {"ATG":0},
            "T": {"ACT":0, "ACC":0, "ACA":0, "ACG":0},
            "N": {"AAT":0, "AAC":0},
            "K": {"AAA":0, "AAG":0},
            "V": {"GTT":0, "GTC":0, "GTA":0, "GTG":0},
            "A": {"GCT":0, "GCC":0, "GCA":0, "GCG":0},
            "D": {"GAT":0, "GAC":0},
            "E": {"GAA":0, "GAG":0},
            "G": {"GGT":0, "GGC":0, "GGA":0, "GGG":0},
            "*": {"TAA":0, "TAG":0, "TGA":0},
        }
        self.count = 0
        self.codonFreq = {k:{} for k in self.codonUsage.keys()}
        self._read_file(input) if type(input) == str else self._read_StealthGenome(input) 
        self._calculate()

    def _read_file(self, input: str) -> None:
        freader = FastAreader.FastAreader(input)
        for _,cds in freader.readFasta():
            cds = cds.seq
            for i in range(len(cds),3):
                codon = input[i:i+3]
                codon_to_aa = self.dnaCodonTable[codon]
                self.codonUsage[codon_to_aa][codon] += 1
    
    def _read_StealthGenome(self,input: SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord]) -> None:
        input = input.cds_sequence if type(input) == SeqParser.StealthGenome else input
        for cds in input:
            cds = cds.seq
            for i in range(0,len(cds),3):
                codon = cds[i:i+3]
                codon_to_aa = self.dnaCodonTable[codon]
                self.codonUsage[codon_to_aa][codon] += 1
                
    def _calculate(self):    
        for aa,codon_dict in self.codonUsage.items():
            aa_sum = sum(codon_dict.values())
            self.count =+ aa_sum
            for codon in codon_dict.keys():
                if aa_sum == 0:
                    self.codonFreq[aa][codon] = 1/len(codon_dict)
                else:
                    self.codonFreq[aa][codon] = self.codonUsage[aa][codon]/aa_sum
                
    def _format_aa_arg(self,aa: str):
        if type(aa) != str:
            raise TypeError(f"Expected {type(str())} got {type(aa)}")
        aa = aa.upper()
        if len(aa) != 1 or aa not in self.validAA:
            raise InvalidAA(f"Invalid single-letter Amino Acid: {aa}")
        return aa
        
                    
    def addCDS(self,cds: str) -> None:
        for i in range(len(cds),3):
            codon = cds[i:i+3]
            codon_to_aa = self.dnaCodonTable[codon]
            self.codonUsage[codon_to_aa][codon] += 1
        self._calculate()
    
    def len(self) -> int:
        return self.count
    
    def getFrequency(self, aa: str = None) -> dict[str,dict[str,float]] | dict[str,float]:
        if aa == None:
            return self.codonFreq
        aa = self._format_aa_arg(aa)
        return self.codonFreq[aa]
    
    def getUsage(self, aa: str = None) -> dict[str,dict[str,int]] | dict[str,int]:
        if aa == None:
            return self.codonUsage
        aa = self._format_aa_arg(aa)
        return self.codonUsage[aa]
       
class CodonOptimizer:
    """
    This is a basic codonOptimizer that takes codon useage data and a sequence
    of aminoAcids as input and recapitulates the overall frequency of the
    given input in the reverse-transcription of the AA seq. This should
    roughly fit to levels of tRNA availibility.
    """

    def __init__(self, frequencies: dict):
        self.synFreqDic = frequencies

    def _weightedChoice(self, aa):
        codons = list(self.synFreqDic[aa].keys())
        frequencies = self.synFreqDic[aa].values()
        chosenCodon = random.choices(codons, weights=frequencies, k=1)[0]
        return chosenCodon

    def assembleSeed(self, aaSequence):
        dnaList = [self._weightedChoice(aa) for aa in aaSequence]
        dnaSeed = "".join(dnaList)
        return dnaSeed
    
    def getFrequency(self) -> dict[str,dict[str,float]] :
        return self.synFreqDic
     
__all__ = ["CDSanalyzer","CodonOptimizer", "InvalidAA"]



"""
TABI - UCSC iGEM 2023

StealthGenome: Prep Genome for Stealth and Codon Frequency analysis

PlasmidParser: Identifies Stealth-mutable regions of an annotated plasmid

Author: Tyler Gaw
"""
# Usage to be written, don't wanna rn

from Bio import SeqIO
from Bio.SeqFeature import SimpleLocation
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from typing import Iterator,Literal
from chameleontools.ORFfinder import ORFfinder
from chameleontools.FastAreader import FastAreader
import sys


class StealthGenome:
    """
    Reads genome file in genbank or FASTA format,

    Stores Genome sequence and CDS regions
    """

    def _validArg(self, genome: str):
        "helper func: checks valid args"
        g_valid = False, False, False
        g_valid = (
            genome.endswith(".fasta")
            or genome.endswith(".gb")
            or genome.endswith(".gbk")
        )
        return g_valid

    def _fastaGenome(self, fasta: str):
        reader = FastAreader(fasta)
        CDS_records = []
        sequence = []
        for frag, (_, seq) in enumerate(reader.readFasta()):
            genome = Seq(seq)
            orfs = ORFfinder(
                seq, longestGene=True, minGene=200
            ).get_genes()  # array of estimated genes
            for i, j in enumerate(orfs):
                "Converting ORFfinder output to Biopython SimpleLocation"
                loc = SimpleLocation(j[0] - 1, j[1], strand=1 if j[3] > 0 else -1)

                "Building Biopython SeqRecord for CDS"
                rec = SeqRecord(loc.extract(genome))
                rec.description = "Predicted Protein Coding Sequence"
                rec.name = "Predicted CDS HEADER"
                rec.id = f"FRAG_{frag} CDS_{i} for CODON STATS"

                CDS_records.append(rec)
            sequence.append(genome)
        return CDS_records, sequence

    ############################################################################################
    # ^^^ Private Helper Functions ^^^
    ############################################################################################

    def __init__(self, genome_infile) -> None:
        """
        Constructor: Validates inputs, writes to files in FASTA format, saves file names

        Can be used for Stealth_CDS_Extract to produce a CDS generator for input file

        Args:
        genome_infile (str): genome file in genbank or FastA format

        Class Attr.:

        self.getGenome() -> array of Seq() 

        self.getCDS() -> Iterator of Biopython SeqRecord for each CDS | list of SeqRecord if using FastA file
        
        self.filetype() -> type of input file | "fasta" if FastA file, "genbank" if GenBank file
        """
        if genome_infile != None:
            g_bool = self._validArg(genome_infile)

            if not g_bool:
                print(
                    f"Genome File needs to be FASTA file (.fasta) or GenBank file (.gb || .gbk). Got {genome_infile}",
                    file=sys.stderr,
                )
                exit()

            if genome_infile.endswith(".fasta"):
                self.cds_sequence, self.genome_sequence = self._fastaGenome(
                    genome_infile
                )
                self.input = "fasta"
            else:
                self.cds_sequence = self.Stealth_CDS_Extract(genome_infile)
                self.genome_sequence = self.genome_sequence
                self.input = "genbank"

    def Stealth_CDS_Extract(self, gbfile: str) -> Iterator[SeqRecord]:
        """
        Reads a genbank file and yields CDS/ORF features for Codon Optimization

        A generator that yields CDS features in FASTA format

        Used to produce data for codon usage statistics
        """
        gb = SeqIO.parse(gbfile, "genbank")
        count = 0
        for rec in gb:
            self.genome_sequence = [rec.seq]  # define self.genome_sequence from genbank record
            for feature in rec.features:
                if feature.type == "CDS" or feature.type == "ORF":
                    count += 1
                    out = feature.extract(rec)

                    """Set the description to 'product' qualifier if avaiable. Else use 'label' qualifier or [null] if labels are not present"""
                    out.description = feature.qualifiers.get(
                        "product", feature.qualifiers.get("label", ["null"])
                    )[0]

                    out.name = "TEMP CDS HEADER"
                    out.id = f"CDS_{count} CODON STATS"
                    yield out
            return  # Only accepts single entry genbank records

    def filetype(self) -> Literal["fasta","genbank"]:
        return self.input
    
    def getGenome(self) -> list[Seq]:
        return self.genome_sequence
    
    def getCDS(self) -> Iterator[SeqRecord] | list[SeqRecord]:
        return self.cds_sequence
        
class PlasmidParse:
    """
    Reads in an annotated plasmid file in genbank format,
    parses mutable regions from CDS/ORF annotations,
    takes statistics for total mutable range
    """

    def _validArg(self, plasmid_infile: str) -> bool:
        "helper func: checks for valid input"
        return plasmid_infile.endswith(".gb") or plasmid_infile.endswith(".gbk")

    def _parsePlasmid(
        self, gbfile: str
    ) -> tuple[list[SimpleLocation], list[SimpleLocation]]:
        "helper func: Filters for CDS and noncoding sequence"
        avoid, cds = [], []
        gb = SeqIO.parse(gbfile, "genbank")
        for rec in gb:
            self.record = rec
            for feature in rec.features:
                if feature.type in {"source", "gene"}:
                    continue
                if feature.type in {"CDS", "ORF"}:
                    cds.append((feature.location, feature.location.start % 3))
                else:
                    avoid.append((feature.location, feature.location.start % 3))
            break  # Only handles single entry genbank records
        return avoid, cds

    def _trimStart(
        self, loc_list: list[SimpleLocation], cds: SimpleLocation
    ) -> list[SimpleLocation]:
        "helper func: trims the start codon of CDS, handles fwd/rev strands"
        ref = (
            SimpleLocation(cds.start, cds.start + 3, cds.strand)
            if cds.strand > 0
            else SimpleLocation(cds.end - 3, cds.end, cds.strand)
        )
        for i in range(len(loc_list)):
            loc = loc_list[i]
            if loc.start < ref.start and loc.end > ref.end:
                loc_list[i] = SimpleLocation(loc.start, ref.start, loc.strand)
                loc_list.insert(i + 1, SimpleLocation(ref.end, loc.end, loc.strand))
                break
            if loc.start in ref:
                loc_list[i] = SimpleLocation(ref.end, loc.end, loc.strand)
                break
            elif loc.end in ref or loc.end == ref.end:
                loc_list[i] = SimpleLocation(loc.start, ref.start, loc.strand)
                break
        return

    def _frameAdj(self, loc_list: list[SimpleLocation], frame: int) -> SimpleLocation:
        "helper func: trims mutable regions to lie in frame"
        new = []
        for loc in loc_list:
            st = loc.start
            ed = loc.end
            for _ in range(1, 3):
                if st % 3 != frame:
                    st += 1
                if ed % 3 != frame:
                    ed -= 1
            if (ed - st) < 3:
                continue
            adj = SimpleLocation(st, ed, loc.strand)
            new.append(adj)
        return new

    def _removeOverlap(
        self, loc_list: list[SimpleLocation], cds: SimpleLocation
    ) -> list[SimpleLocation]:
        "helper func: removes overlapping cds regions"
        ret = []
        for loc in loc_list:
            new = [[loc.start, loc.end]]
            if loc.start in cds and (loc.end in cds or loc.end == cds.end):
                "completely removes region"
                continue
            if loc.start < cds.start and loc.end > cds.end:
                "splits range into two"
                dummy = [cds.end, loc.end]
                new[-1][1] = cds.start
                new.append(dummy)
            elif loc.start in cds:
                "truncates start"
                new[-1][0] = cds.end
            elif loc.end in cds or loc.end == cds.end:
                "truncates end"
                new[-1][1] = cds.start + 1
            ret.extend([SimpleLocation(i, j, loc.strand) for i, j in new])
        return ret

    def _windowCorrection(
        self,
        old: list[list[SimpleLocation], int, SimpleLocation],
        st_bound: int,
        ed_bound: int,
    ) -> list[list[SimpleLocation], int, SimpleLocation]:
        "helper func: corrects for out-frame overlapping CDS"
        loc, fm, parent = old
        new = SimpleLocation(st_bound, ed_bound, parent.strand)
        dummy = []

        for i in loc:
            if st_bound > i.end:
                continue
            if i.start in new and (i.end in new or i.end == new.end):
                dummy.append(i)
            elif i.start in new:
                dummy.append(SimpleLocation(i.start, ed_bound, i.strand))
                continue
            elif i.end in new or i.end == new.end:
                dummy.append(SimpleLocation(st_bound, i.end, i.strand))
                continue

        return [dummy, fm, new]

    ############################################################################################
    # Private Helper Functions
    ############################################################################################

    def __init__(self, plasmid_infile: str) -> None:
        """
        Constructor: Validates inputs, parses mutable regions

        Args:
        plasmid_infiled (str): genome file in FASTA or genbank format

        Class Methods:
        
        self.getGenBank() -> Biopython SeqRecord() of input plasmid, contains feature annotations

        self.getSeq() -> Biopython Seq() of input Plasmid

        self.regions() -> array of Biopython SimpleLocation() ranges of mutable codons

        self.mutableCount() -> final mutable regions on plasmid

        self.unmutableCount() -> number of removed basepairs from total CDS coverage

        self.regionCount() -> Total CDS coverage before extracting mutable regions
        """
        p_bool = self._validArg(plasmid_infile)
        if not p_bool:
            print(
                f"Plasmid Input File needs to be GenBank file (.gb || .gbk). Got: {plasmid_infile}",
                file=sys.stderr,
            )
            exit()

        "Did this cause it was ugly when using tuple unpacking"
        temp_parse = self._parsePlasmid(plasmid_infile)

        temp = self.defineMutable(temp_parse)
        
        self.record = self.record 
        
        self.plasmid_sequence = self.record.seq # self.record -> Bio.SeqRecord of plasmid

        self.mutable_regions = temp[0]
        self.total_mutable = temp[1]
        self.removed = temp[2]

        self.total_coverage = self.total_mutable + self.removed

    def defineMutable(
        self, location: tuple[list[SimpleLocation]]
    ) -> list[SimpleLocation]:
        "Finds non-overlapped mutable regions in an annotated plasmid"

        avoid, cdsRegions = location
        temp = []  # [[[SimpLocation List],frame,parent CDS]]

        "Removes mutable cds regions from overlapping non-coding sequence"
        for cds, frame in cdsRegions:
            if len(cds) <= 3:
                continue
            bounds = [[cds.start, cds.end]]
            for loc, _ in avoid:
                if cds.start > loc.start and cds.end < loc.end:
                    "case: interior ORF to non-coding region"
                    bounds = [[-1, -1]]
                    break
                if bounds[-1][0] < loc.start and bounds[-1][1] > loc.end:
                    "case: non-coding region interior to ORF"
                    new_bound = [loc.end, cds.end]
                    bounds[-1][1] = loc.start
                    bounds.append(new_bound)
                elif bounds[-1][0] in loc:
                    "case: clipped start"
                    bounds[-1][0] = loc.end
                elif bounds[-1][1] in loc or bounds[-1][1] == loc.end:
                    "case: clipped end"
                    bounds[-1][1] = loc.start
            if (bounds[-1][1] - bounds[-1][0]) <= 3:
                del bounds[-1]
            if len(bounds) <= 0:
                continue
            mut_range = [
                SimpleLocation(start=rng[0], end=rng[1], strand=cds.strand)
                for rng in bounds
            ]
            temp.append([mut_range, frame, cds])

        "Trimming start codons"
        for i in range(len(temp)):
            tLoc = temp[i][0]
            frm = temp[i][1]
            reg = temp[i][2]
            window = temp[i + 1 :]
            for j in range(len(window)):
                _, parent_frame, del_cds = window[j]
                if (del_cds.start in reg and del_cds.strand > 0) or (
                    del_cds.end in reg and del_cds.strand < 0
                ):
                    self._trimStart(temp[i][0], del_cds)
                    tLoc = temp[i][0]
            if reg.strand > 0:
                trimmed = SimpleLocation(reg.start + 3, reg.end, reg.strand)
                if tLoc[0].start == reg.start:
                    temp[i][0][0] = SimpleLocation(
                        trimmed.start, tLoc[0].end, tLoc[0].strand
                    )
            else:
                trimmed = SimpleLocation(reg.start, reg.end - 3, reg.strand)
                if tLoc[-1].end == reg.end:
                    temp[i][0][-1] = SimpleLocation(
                        tLoc[-1].start, trimmed.end, tLoc[-1].strand
                    )

        "sorts regions by CDS length, compares larger CDS to smaller CDS to remove overlap"
        temp = sorted(temp, key=lambda x: len(x[2]), reverse=True)

        "remove overlaps"
        for i in range(len(temp)):
            frag = temp[i][0]
            frm = temp[i][1]
            region = temp[i][2]
            window = temp[i + 1 :]
            if region == None:
                continue
            for j in range(len(window)):
                _, parent_frame, parent_cds = window[j]
                if parent_cds == None:
                    continue
                if parent_cds.start in region or parent_cds.end in region:
                    if parent_frame != frm:
                        temp[i][0] = self._removeOverlap(frag, parent_cds)
                        frag = temp[i][0]
                if parent_cds.start in region and parent_cds.end in region:
                    temp[i + j + 1][2] = None
                elif parent_cds.start in region:
                    temp[i + j + 1] = self._windowCorrection(
                        window[j], region.end, parent_cds.end
                    )
                elif parent_cds.end in region:
                    temp[i + j + 1] = self._windowCorrection(
                        window[j], parent_cds.start, region.start
                    )

        "frame adjusts mutable fragments to lie in frame of parent CDS"
        for i in range(len(temp)):
            temp[i][0] = self._frameAdj(temp[i][0], temp[i][1])

        mutable_regions = sorted(
            [region for CDS, _, tf in temp for region in CDS if tf is not None],
            key=lambda x: x.start,
        )

        "mutable region statistics - total covered region and mutable regions"
        total_coverage = sum([len(x[-1]) for x in temp if x[-1] is not None])
        removed = total_coverage - sum([len(x) for x in mutable_regions])
        return mutable_regions, total_coverage, removed

    def getSeq(self) -> Seq:
        return self.plasmid_sequence
    
    def regions(self) -> list[SeqRecord]:
        return self.mutable_regions
    
    def mutableCount(self) -> int:
        return self.total_mutable
    
    def regionCount(self) -> int:
        return self.total_coverage
    
    def unmutableCount(self) -> int:
        return self.removed 
    
    def getGenBank(self) -> SeqRecord:
        return self.record
    

__all__ = ["StealthGenome","PlasmidParse"]


"""
TABI - UCSC iGEM 2023
Modified from ORFfinder Class written for Winter 2022 BME160
BME160 Professor: David L. Bernick
"""

# ORF finder used to parse FastA format genome files.
# Optional parameters:
#     longestGene -> only reports longest genes in an ORF
#     minGene -> only reports genes of this size or larger
#     starts -> define set of valid start codons
#     stop -> define set fo valid stop codons

from ._ORFfinder import *


"""
TABI - UCSC iGEM 2023
Modified from ORFfinder Class written for Winter 2022 BME160
BME160 Professor: David L. Bernick
"""

# ORF finder used to parse FastA format genome files.
# Optional parameters:
#     longestGene -> only reports longest genes in an ORF
#     minGene -> only reports genes of this size or larger
#     starts -> define set of valid start codons
#     stop -> define set fo valid stop codons

class ORFfinder:
    """
    ORF finder class that scans through a genome and reports ORFs from all frames

    Data is reported in a list of lists in

    format: list[[startPos,stopPos,length,frame]]
    """

    def __init__(
        self,
        seq: str,
        longestGene=False,
        minGene=100,
        starts: set = {"ATG"},
        stops: set = {"TAA", "TAG", "TGA"},
    ):
        """_summary_

        Args:
            seq (str): _description_
            longestGene (bool, optional): _description_. Defaults to False.
            minGene (int, optional): _description_. Defaults to 100.
            starts (set, optional): _description_. Defaults to {"ATG"}.
            stops (set, optional): _description_. Defaults to {"TAA", "TAG", "TGA"}.
        """
        assert minGene >= 0
        self.geneCandidates = []
        format = seq.strip().upper().replace(" ", "")
        flipTable = format[::-1].maketrans(
            "ATCG", "TAGC"
        )  # Builds reverse compliment strand

        self.fiveThree = format
        self.threeFive = format[::-1].translate(flipTable)
        self.validStart = starts
        self.validStops = stops
        self.minGene = minGene
        self.longestGene = longestGene

        self._analysis()

    def _geneFinder(self, seq, rev, longGene):
        """
        Helper function that collects starts and identifies stops
        """
        start = set()
        for frame in range(0, 3):
            start.clear()
            for i in range(frame, len(seq), 3):
                codon = seq[i : i + 3]

                if codon in self.validStart:
                    start.add(i)

                if codon in self.validStops and len(start) > 0:
                    _frame = -1 - frame if rev else frame + 1
                    self._addGene(start, i, _frame, longGene)
                    start = set()

            if len(start) > 0:
                _frame = -1 - frame if rev else frame + 1
                self._addGene(start, len(seq) - 3, _frame, longGene)

    def _addGene(self, start, i, frame, longGene):
        """
        Helper Function
        Does all gene-candidate calculations using [start] [stop index] [frame #] [longest gene flag]
        """
        for s in sorted(start):
            length = i - s + 3
            if frame < 0:  # Handles reverse strand calcs
                startPos = len(self.fiveThree) - i - 2
                stopPos = len(self.fiveThree) - s
            else:  # Handles top strand
                startPos = s + 1
                stopPos = i + 3
            if length >= self.minGene:
                self.geneCandidates.append([startPos, stopPos, length, frame])
            if longGene:  # LongestGene flag
                return

    def _analysis(self):
        """
        Main geneFinder function. Runs hidden _geneFinder() function to parse through both top/bottom strand
        """
        self._geneFinder(self.fiveThree, False, self.longestGene)
        self._geneFinder(self.threeFive, True, self.longestGene)

        # sorts geneCandidates list, sorting by gene length
        self.formatted = [
            x
            for x in sorted(
                self.geneCandidates, key=lambda entry: entry[2], reverse=True
            )
        ]

    def get_genes(self) -> list[tuple[int,int,int,int]]:
        return self.geneCandidates

__all__ = ["ORFfinder"]


import argparse
import random
from Bio.Seq import Seq


class codonOptimizer:
    """
    This is a basic codonOptimizer that takes codon useage data and a sequence
    of aminoAcids as input and recapitulates the overall frequency of the
    given input in the reverse-transcription of the AA seq. This should
    roughly fit to levels of tRNA availibility.
    """

    def __init__(self, synonomous, frequencies):
        self.synFreqDic = {}
        for aa, codons in synonomous.items():
            codFreqList = [(codon, frequencies[codon]) for codon in codons]
            self.synFreqDic[aa] = codFreqList

    def weightedChoice(self, aa):
        codons, frequencies = zip(*self.synFreqDic[aa])
        chosenCodon = random.choices(codons, weights=frequencies, k=1)
        return chosenCodon[0]

    def assembleSeed(self, aaSequence):
        dnaList = [self.weightedChoice(aa) for aa in aaSequence]
        dnaSeed = "".join(dnaList)
        return dnaSeed


class patternConstrainer:
    """
    Initialized via an instance of putativly "optimized" sequence,
    patternConstrainer finds regions of the seed that contain motifs of
    interest (via a 5'->3' sliding window approach) and passes associated
    indicies (start pos. based on specific window size & fixed adapter
    sequences at the begainning and end of each window) to some outside
    func.
    In this case, the outside func returns a 'more ideal' (synonomous variate)
    which patternConstrainer catches and overwrites into the working sequence.

    edge case handling:
    """

    def __init__(self, motifs, codonOptions, codonPref):
        self.motifs = motifs  # A list of motifs to avoid, likely to trigger RMS
        self.codonOptions = codonOptions  # The dictionary of codons mapping to their corresponding amino acids.
        self.codonPref = codonPref

    def motifInSeq(self, sequence):
        """
        remove? if we use a motif checker that checks for a motif at each position, we can add variation to the shuffles actual footprint size
        """
        # Iterate over each motif
        for motif in self.motifs:
            # If the motif is found in the sequence
            if motif in sequence:
                return True  # Return True immediately. This indicates that the sequence contains a motif that we want to avoid.
        return False  # If none of the motifs were found, return False. This indicates that the sequence is safe to use as is.

    def optimizeCodon(self, sequence, start):
        length = len(sequence)
        workingDnaSequence = sequence[start : start + 12]

        bioSeq = Seq(workingDnaSequence)
        workingAaSeq = bioSeq.translate()

        if start < 9:
            headStart = start
        elif 9 <= start:
            headStart = start - 9

        if length <= start + 21:
            tailEnd = length
            tailStart = length
        else:
            tailEnd = start + 21
            tailStart = start + 12

        headAdapt = sequence[
            headStart:start
        ]  # change to handle begainning case (no -6 index)
        tailAdapt = sequence[tailStart:tailEnd]

        myPermuter = Permutations(self.motifs, self.codonOptions, self.codonPref)
        cleanedRegion = myPermuter.rank(
            workingDnaSequence, workingAaSeq, headAdapt, tailAdapt
        )

        cleanCutOff = len(cleanedRegion) - len(tailAdapt)
        newSequence = (
            sequence[:start]
            + cleanedRegion[len(headAdapt) : cleanCutOff]
            + sequence[start + 12 :]
        )

        return newSequence

    def optimizeSequence(self, sequence):
        """
        understand behaviour
        """
        optimizedSequence = sequence
        for i in range(9, len(sequence) - 9, 3):
            if self.motifInSeq(optimizedSequence[i : i + 12]):
                newSequence = self.optimizeCodon(optimizedSequence, i)
                optimizedSequence = newSequence
        return optimizedSequence


class Permutations:
    """
    Generates all possible codon permutations of given window (codon subsequence
    of sequence) & ranks permutations based on rank items as they are generated.
    Each round of generation ends with the appension of scores and permutations
    added to net list as a tuple. These tuples are then sorted via score descrimination-
    only canditates with the best score for first item contenue, same with second item
    and then the 3rd item varies. This can be adjusted but allows for the prioritization
    of sorting methodologies for each specific score items.
    prior implementations had score bleeding issues, this is currently an open end
    """

    def __init__(self, motifs, codonOptions, frequencies):
        self.Motifs = motifs
        self.frequencies = frequencies
        self.codonOptions = codonOptions

    def permute(self, amino_acid_sequence, current_codons=""):
        if not amino_acid_sequence:
            yield current_codons
            return

        current_aa = amino_acid_sequence[0]
        remaining_aas = amino_acid_sequence[1:]

        for codon in self.codonOptions[current_aa]:
            yield from self.permute(remaining_aas, current_codons + codon)

    def rank(self, dnaSeq, aaSeq, headAdapt, tailAdapt):
        """ """
        ranks = []
        permutation_generator = self.permute(aaSeq)

        while True:
            try:
                putative = next(permutation_generator)
                permutation = headAdapt + putative + tailAdapt

                severity, codScore = self.analyze_dna_motifs(permutation)
                adherance = self.gaugeAdherance(
                    dnaSeq, putative
                )  # returns a value between 0 and 1, 0 being less adherant

                ranks.append((severity, adherance, codScore, permutation))
            except StopIteration:
                break  # Stop when all permutations have been generated

        min_first_item = min(ranks, key=lambda x: x[0])[0]
        filter1 = sorted(
            [t for t in ranks if t[0] == min_first_item],
            key=lambda x: x[1],
            reverse=True,
        )  # Sort the remaining tuples based on the second item in descending order

        max_second_item = filter1[0][1]
        filter2 = [
            t for t in filter1 if t[1] == max_second_item
        ]  # Extract tuples with the highest second item values

        max_third_item = max(filter2, key=lambda x: x[2])[2]
        filter3a = [
            t for t in filter2 if 0.95 * max_third_item <= t[2] <= 1.05 * max_third_item
        ]  # Extract tuples with a third item value within 5% of the maximal value
        filter3b = sorted(filter3a, key=lambda x: x[2], reverse=True)

        # add another layer, for now first item
        chosenTup = filter3b[0]
        chosenOne = chosenTup[3]
        return chosenOne

    def analyze_dna_motifs(self, sequence):
        occurrences = {}  # Dictionary to store occurrences of short motifs
        total_occurrences = 0  # Total count of occurrences in the longer sequence

        for short_motif in self.Motifs:
            motif_positions = []  # List to store positions of motif occurrences
            motif_count = 0  # Count of occurrences for each motif
            start_position = 0

            while start_position < len(sequence):
                position = sequence.find(short_motif, start_position)

                if position == -1:
                    break  # No more occurrences found
                motif_positions.append(position)
                motif_count += 1
                total_occurrences += 1
                start_position = position + 1  # Move to the next position

            occurrences[short_motif] = {
                "positions": motif_positions,
                "count": motif_count,
            }

        score = 0.0
        codLen = len(sequence) / 3
        for i in range(0, len(sequence), 3):
            codon = sequence[i : i + 3]
            if codon in self.frequencies:
                score += self.frequencies[codon]
                Score = score / codLen
        return total_occurrences, Score

    def gaugeAdherance(self, refSeq, perm):
        codonRef = []
        codonPerm = []
        adherance = 0
        """
        for i in range(0, len(refSeq), 3): #both refSeq & perm should be of the same length
            codon = refSeq[ i : i + 3 ]
            codonRef.append(codon)
        for i in range(0, len(perm), 3):
            codon = perm[ i : i + 3 ]
            codonPerm.append(codon)
        """
        for i in range(0, len(refSeq), 3):
            refCodon = refSeq[i : i + 3]
            codonRef.append(refCodon)

            perCodon = perm[i : i + 3]
            codonPerm.append(perCodon)

        adherance = 0
        codLen = len(refSeq) / 3
        for tup in zip(codonRef, codonPerm):
            cod1 = tup[0]
            cod2 = tup[1]
            if cod1 == cod2:
                adherance += 1 / codLen
        return adherance  # confers percentage of conserved codons in perm


class MotifChecker:
    """
    helper class, used to fetch final stats in main
    """

    def __init__(self, motifs):
        self.motifs = motifs

    def checkMotifs(self, sequence):
        totalOccurrences = 0  # Total count of occurrences in the longer sequence

        for motif in self.motifs:
            startPosition = 0
            while startPosition < len(sequence):
                position = sequence.find(motif, startPosition)

                if position == -1:
                    break  # No more occurrences found

                totalOccurrences += 1
                startPosition = position + 1  # Move to the next position

        return totalOccurrences


def main():
    # codonOptions is a constant
    codonOptions = {
        "F": ["TTT", "TTC"],
        "L": ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"],
        "S": ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"],
        "Y": ["TAT", "TAC"],
        "C": ["TGT", "TGC"],
        "W": ["TGG"],
        "P": ["CCT", "CCC", "CCA", "CCG"],
        "H": ["CAT", "CAC"],
        "Q": ["CAA", "CAG"],
        "R": ["CGT", "CGC", "CGA", "CGG", "AGA", "AGG"],
        "I": ["ATT", "ATC", "ATA"],
        "M": ["ATG"],
        "T": ["ACT", "ACC", "ACA", "ACG"],
        "N": ["AAT", "AAC"],
        "K": ["AAA", "AAG"],
        "V": ["GTT", "GTC", "GTA", "GTG"],
        "A": ["GCT", "GCC", "GCA", "GCG"],
        "D": ["GAT", "GAC"],
        "E": ["GAA", "GAG"],
        "G": ["GGT", "GGC", "GGA", "GGG"],
        "*": ["TAA", "TAG", "TGA"],
    }

    # Create a parser with appropriate descriptions for the arguments
    parser = argparse.ArgumentParser(description="DNA Processing")
    parser.add_argument("amino_sequence", help="Amino sequence as a string")
    parser.add_argument("frequencies", help="Frequencies as a dictionary string")
    parser.add_argument("motifs", help="Motifs as a list string")

    args = parser.parse_args()

    # Parse the input arguments
    aaSequence = args.amino_sequence

    # Convert the input strings to Python dictionaries and lists
    frequencies = eval(args.frequencies)  # Use eval to convert to a dictionary
    motifs = eval(args.motifs)  # Use eval to convert to a list

    """
    frequencies = {'TAA': 0.526, 'TAG': 0.264, 'TGA': 0.21, 'GCA': 0.182, 'GCC': 0.316, 
                    'GCG': 0.168, 'GCT': 0.334, 'TGC': 0.259, 'TGT': 0.741, 'GAC': 0.233,
                    'GAT': 0.767, 'GAA': 0.741, 'GAG': 0.259, 'TTC': 0.242, 'TTT': 0.758,
                    'GGA': 0.292, 'GGC': 0.184, 'GGG': 0.182, 'GGT': 0.343, 'CAC': 0.351,
                    'CAT': 0.649, 'ATA': 0.128, 'ATC': 0.324, 'ATT': 0.548, 'AAA': 0.791,
                    'AAG': 0.209, 'CTA': 0.126, 'CTC': 0.129, 'CTG': 0.117, 'CTT': 0.099,
                    'TTA': 0.389, 'TTG': 0.14, 'ATG': 1.0, 'AAC': 0.277, 'AAT': 0.723,
                    'CCA': 0.162, 'CCC': 0.411, 'CCG': 0.172, 'CCT': 0.256, 'CAA': 0.695,
                    'CAG': 0.305, 'AGA': 0.19, 'AGG': 0.067, 'CGA': 0.153, 'CGC': 0.218, 
                    'CGG': 0.139, 'CGT': 0.234, 'AGC': 0.142, 'AGT': 0.266, 'TCA': 0.107,
                    'TCC': 0.173, 'TCG': 0.103, 'TCT': 0.211, 'ACA': 0.18, 'ACC': 0.36, 
                    'ACG': 0.139, 'ACT': 0.32, 'GTA': 0.181, 'GTC': 0.227, 'GTG': 0.264, 
                    'GTT': 0.328, 'TGG': 1.0, 'TAC': 0.283, 'TAT': 0.717}
    motifs = ['CACCTGC', 'GTGGACG', 'CCGG', 'CGCG', 'GGCC', 'TGCA', 'CCAGG', 'CCTGG', 'GGACC', 'GGCCC', 'GGGCC', 'GGTCC', 'TCTGA', 'AAATTT', 'AATATT', 'ACATGT', 'ACGCGT', 'ACTAGT', 'AGCGCT', 'AGTACT', 'ATCGAT', 'ATGCAT', 'CAATTG', 'CAGCTG', 'CATATG', 'CCGCGG', 'CCTAGG', 'CGCGCG', 'CGTACG', 'CTCGAG', 'CTGCAG', 'GAATTC', 'GACGTC', 'GAGCTC', 'GATATC', 'GCATGC', 'GCGCGC', 'GCTAGC', 'GGCGCC', 'GGTACC', 'GTATAC', 'GTCGAC', 'GTGCAC', 'TAATTA', 'TACGTA', 'TCCGGA', 'TGCGCA', 'TGTACA', 'TTCGAA', 'TTTAAA', 'GGCC', 'ACTAGT', 'AGTACT', 'ATCGAT', 'ATGCAT', 'CTCGAG', 'CTGCAG', 'GAATTC', 'GCTAGC', 'GGTACC', 'GTCGAC', 'TAATTA', 'TGCGCA', 'TTCGAA', 'CCTCAGG', 'CCTGAGG', 'GCTAAGC', 'GCTTAGC', 'GGTAACC', 'GGTCACC', 'GGTGACC', 'GGTTACC', 'AAAATTTT', 'AAACGTTT', 'AAAGCTTT', 'AAATATTT', 'AACATGTT', 'AACCGGTT', 'AACGCGTT', 'AACTAGTT', 'AATATATT', 'AATCGATT', 'AATGCATT', 'AATTAATT', 'AGATATCT', 'AGTTAACT', 'ATAATTAT', 'ATTATAAT', 'CAGATCTG', 'CAGTACTG', 'CATATATG', 'CATTAATG', 'CCAATTGG', 'CCACGTGG', 'CCAGCTGG', 'CCCATGGG', 'CCCCGGGG', 'CCCGCGGG', 'CGAGCTCG', 'CGATATCG', 'CGCGCGCG', 'CGCTAGCG', 'CTAATTAG', 'CTAGCTAG', 'CTATATAG', 'CTGATCAG', 'CTGGCCAG', 'CTGTACAG', 'CTTATAAG', 'CTTGCAAG', 'CTTTAAAG', 'GAGATCTC', 'GAGTACTC', 'GATATATC', 'GATTAATC', 'GTGATCAC', 'GTTATAAC', 'TACCGGTA', 'TACTAGTA', 'TATCGATA', 'TATTAATA', 'TGATATCA', 'TGTTAACA']
    aaSequence = 'MVSKGEELFTGVVPILVELDGDVNGHKFSVSGEGEGDATYGKLTLKFICTTGKLPVPWPTLVTTLTYGVQCFSRYPDHMKQHDFFKSAMPEGYVQERTIFFKDDGNYKTRAEVKFEGDTLVNRIELKGIDFKEDGNILGHKLEYNYNSHNVYIMADKQKNGIKVNFKIRHNIEDGSVQLADHYQQNTPIGDGPVLLPDNHYLSTQSALSKDPNEKRDHMVLLEFVTAAGITLGMDELYK'
    """

    seedDNA = codonOptimizer(codonOptions, frequencies)
    avoidPatterns = patternConstrainer(motifs, codonOptions, frequencies)
    evalMortality = MotifChecker(motifs)

    roundsData = []
    i = 0
    while i < 5:
        seed = seedDNA.assembleSeed(aaSequence)
        optimization = avoidPatterns.optimizeSequence(seed)

        mortality = evalMortality.checkMotifs(optimization)
        roundsData.append((mortality, optimization))
        i += 1
    lowestMortality = min(roundsData, key=lambda x: x[0])
    best = lowestMortality[0]
    bestSeq = lowestMortality[1]

    print(best, "MOTIFS REMAIN: \n", bestSeq)


if __name__ == "__main__":
    main()



from ._stealth import Genome,rc as reverse_complement
from chameleontools.SeqParser import StealthGenome
from Bio.Seq import Seq

class StealthV0:
    
    def __init__(self,seq: Seq | str | list[str | Seq] | StealthGenome , zScore_cut: float, pseudo: int, kMax: int, kMin: int = 1) -> None:
        if type(seq) == StealthGenome:
            seq = seq.getGenome()
        self.output = self._runStealth(seq, zScore_cut, pseudo, kMax, kMin)
    
    def _runStealth(self,genome: str, zScore_cut: float, pseudo: int, kMax: int, kMin: int = 1) -> list[tuple[str,float,bool]]:
        """
        runs Stealth analysis on input genome
        saves output to array -> [(motif,zScore)]
        """
        if type(genome) != list:
            genome = [genome]
        stealth = Genome(1,kMax,pseudo)
        for seq in genome:
            stealth.addSequence(str(seq))# conversion of BioPython Seq() to str 
        """Taken from original stealth.vo.py main() func"""
        ######V Written by David L. Bernick V#######
        outArray = []
        for seq,_ in stealth.counts.items():
            if len(seq) < kMin: continue
            # for odd  len sequence, we just print the composite that is for the full sequence.
            if 'N' in seq: # this is a composite
                p = seq.find('N')
                if len(seq) != (p*2+1): continue # skip this one, it is a part of a composite
            elif len(seq) % 2 == 1: continue # odd sequences print only if they are a composite
        ######^ Written by David L. Bernick ^#######
            zscore = stealth.Zscore(seq)
            if  zscore <= zScore_cut:
                if (seq in stealth.counts):
                    rc_flag = False
                    if reverse_complement(seq):
                        rc_flag = True    
                    outArray.append((seq,zscore,rc_flag))
        return outArray
    
    def getOutput(self) -> list[tuple[str,float,bool]]:
        return self.output
    
__all__ = ["StealthV0"]